;;; imap.el --- imap library
;; Copyright (C) 1998 Free Software Foundation, Inc.

;; Author: Simon Josefsson <jas@pdc.kth.se>
;; This file is not part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;;
;; RFC1730 (IMAP4): done
;; RFC1731 (Authentication mecanisms): currently only support for KERBEROS_V4
;; RFC2060 (IMAP4rev1): done
;; RFC???? (UNSELECT ext): done
;; RFC2195 (CRAM-MD5 auth): done
;; RFC2086 (ACL ext): done
;; RFC2342 (NAMESPACE ext): done
;; RFC2359 (UIDPLUS ext): done
;;
;; Todo:
;; 
;; o Handle literals in LIST/LSUB untagged responses
;; o Fix imap-ok-p.
;; o Parse UIDs as strings? (28 bit limit)
;; o Sleep.
;; 

;;; Code:

(require 'rfc2104)
(require 'base64)
(require 'md5)

(eval-when-compile (require 'cl))

(eval-and-compile
  (autoload 'open-ssl-stream "ssl")
  (unless (fboundp 'open-network-stream)
    (require 'tcp)))

;; User variables.

(defvar imap-imtest-arguments "-kp"
  "Privacy / integrity checking arguments passed to `imtest'.")

(defvar imap-default-user (user-login-name)
  "Default username to use.")

;; Various variables.

(defvar imap-cb-fetch-hook nil
  "Hook called when receiving a FETCH response. Called with article NUM,
FETCH and DATA response.")

(defvar imap-streams '(kerberos4 ssl network)
  "Priority of streams to consider when opening connection to
server.")

(defvar imap-stream-alist
  '((kerberos4 imap-kerberos4s-p imap-kerberos4-open)
    (ssl       imap-ssl-p        imap-ssl-open)
    (network   imap-network-p    imap-network-open))
  "Definition of network streams.

(NAME CHECK OPEN)

NAME names the stream, CHECK is a function returning non-nil if the
server support the stream and OPEN is a function for opening the
stream.")

(defvar imap-authenticators '(kerberos4 cram-md5 login anonymous)
  "Priority of authenticators to consider when authenticating to
server.")

(defvar imap-authenticator-alist 
  '((kerberos4 imap-kerberos4a-p imap-kerberos4-auth)
    (cram-md5  imap-cram-md5-p   imap-cram-md5-auth)
    (login     imap-login-p      imap-login-auth)
    (anonymous imap-anonymous-p  imap-anonymous-auth))
  "Definition of authenticators.

(NAME CHECK AUTHENTICATE)

NAME names the authenticator. CHECK is a function returning non-nil if
the server support the authenticator and AUTHENTICATE is a function
for doing the actuall authentification.")

;; Internal constants. Change theese and die.

(defconst imap-default-port 143)
(defconst imap-default-ssl-port 993)
(defconst imap-default-stream 'network)
(defconst imap-coding-system-for-read 'binary)
(defconst imap-coding-system-for-write 'binary)
(defconst imap-local-variables '(imap-server
				 imap-port
				 imap-client-eol
				 imap-server-eol
				 imap-auth
				 imap-stream
				 imap-username
				 imap-password
				 imap-current-mailbox
				 imap-capability
				 imap-namespace
				 imap-state
				 imap-reached-tag
				 imap-tag
				 imap-process
				 imap-mailbox-data))
(defconst imap-parse-response-data-cb
  '((OK         . imap-response-data-text-code)
    (NO         . imap-response-data-text-code)
    (BAD        . imap-response-data-text-code)
    (BYE        . imap-response-data-bye)
    (EXISTS     . imap-response-data-exists)
    (EXPUNGE    . imap-response-data-expunge)
    (RECENT     . imap-response-data-recent)
    (CAPABILITY . imap-response-data-capability)
    (LIST       . imap-response-data-list)
    (LSUB       . imap-response-data-list)
    (FLAGS      . imap-response-data-flags)
    (FETCH      . imap-response-data-fetch)
    (SEARCH     . imap-response-data-search)
    (STATUS     . imap-response-data-status)
    (ACL        . imap-response-data-acl)
    (NAMESPACE  . imap-response-data-namespace)))


;; Internal variables.

(defvar imap-buffer-list nil
  "List of buffers the imap library has created, which are destroyed
on call to `imap-done'.")

(defvar imap-stream nil)
(defvar imap-auth nil)
(defvar imap-server nil)
(defvar imap-port nil)
(defvar imap-username nil)
(defvar imap-password nil)
(defvar imap-state 'closed 
  "IMAP state. Valid states are `closed', `initial', `nonauth',
`auth', `selected' and `examine'.")

(defvar imap-server-eol "\r\n"
  "The EOL string sent from the server.")

(defvar imap-client-eol "\r\n"
  "The EOL string we send to the server.")

(defvar imap-current-mailbox nil
  "Current mailbox name.")

(defvar imap-mailbox-data nil
  "Obarray with mailbox data.")

(defvar imap-mailbox-prime 127
  "Length of imap-mailbox-data.")

(defvar imap-current-message nil
  "Current message number.")

(defvar imap-message-data nil
  "Obarray with message data.")

(defvar imap-message-prime 67
  "Length of imap-message-data.")

(defvar imap-capability nil
  "Capability for server.")

(defvar imap-namespace nil
  "Namespace for current server.")

(defvar imap-reached-tag 0
  "Lower limit on command tags that have been parsed.")

(defvar imap-tag 0
  "Command tag number.")

(defvar imap-process nil
  "Process.")

(defvar imap-log "*imap-log*"
  "Imap session trace.")

(defvar imap-debug nil;"*imap-debug*"
  "Random debug spew.")


;; Server functions; stream stuff:

(defun imap-kerberos4s-p (buffer)
  (imap-capability 'AUTH=KERBEROS_V4 buffer))

(defun imap-kerberos4-open (name buffer server port)
  (setq port (or port imap-default-port))
  (let* ((coding-system-for-read imap-coding-system-for-read)
	 (coding-system-for-write imap-coding-system-for-write)
	 (process (start-process name buffer "imtest" imap-imtest-arguments
				 server (number-to-string port))))
    (with-current-buffer (process-buffer process)
      (setq imap-client-eol "\n")
      (when process
	(message "Opening Kerberized IMAP connection...")
	;; Result of authentication is a string: __Full privacy protection__
	(while (and (memq (process-status process) '(open run))
		    (goto-char (point-min))
		    (not (and (imap-parse-greeting)
			      (re-search-forward "__\\(.*\\)__\n" nil t))))
	  (accept-process-output process 1)
	  (sit-for 1))
	(and imap-log
	     (with-current-buffer (get-buffer-create imap-log)
	       (goto-char (point-max))
	       (insert-buffer (process-buffer process))))
	(let ((response (match-string 1)))
	  (erase-buffer)
	  (message "Kerberized IMAP connection: %s" response)
	  (if (let ((case-fold-search nil)) 
		(not (string-match "failed" response)))
	      process
	    (imap-send-command-wait "LOGOUT")
	    (delete-process process)
	    nil))))))
  
(defun imap-ssl-p (buffer)
  nil)

(defun imap-ssl-open-1 (name buffer server port extra-arg)
  (setq port (or port imap-default-ssl-port))
  (let* ((coding-system-for-read imap-coding-system-for-read)
	 (coding-system-for-write imap-coding-system-for-write)
	 (ssl-program-arguments (list extra-arg
				      "-connect" 
				      (format "%s:%d" imap-server port)))
	 (proc (open-ssl-stream name buffer server port)))
    (with-current-buffer buffer
      (goto-char (point-min))
      (while (and (memq (process-status proc) '(open run))
		  (goto-char (point-min))
		  (not (re-search-forward "^\r*\\* OK" nil t)))
	(accept-process-output proc 1)
	(goto-char (point-min)))
      (beginning-of-line)
      (delete-region (point-min) (point))
      (when (memq (process-status proc) '(open run))
	proc))))

(defun imap-ssl-open (name buffer server port)
  (message "Opening SSL3 connection...")
  (let ((ret (imap-ssl-open-1 name buffer server port "-ssl3")))
    (if ret
	ret
      (message "Opening SSL2 connection...")
      (imap-ssl-open-1 name buffer server port "-ssl2"))))

(defun imap-network-p (buffer)
  t)

(defun imap-network-open (name buffer server port)
  (setq port (or port imap-default-port))
  (let ((coding-system-for-read imap-coding-system-for-read)
	(coding-system-for-write imap-coding-system-for-write))
    (condition-case ()
	(open-network-stream name buffer server port)
      (error nil)
      (quit nil))))
  
;; Server functions; authenticator stuff:

(defun imap-interactive-login (buffer loginfunc)
  "Login to server in BUFFER. LOGINFUNC is passed a username and a
password, it should return t if it where sucessful authenticating
itself to the server, nil otherwise. Returns t if login was
successful, nil otherwise."
  (with-current-buffer buffer
    (make-variable-buffer-local 'imap-username)
    (make-variable-buffer-local 'imap-password)
    (let (user passwd ret)
      (condition-case ()
	  (while (or (not user) (not passwd))
	    (setq user (or imap-username
			   (read-from-minibuffer 
			    (concat "IMAP username for " imap-server ": ")
			    (or user imap-default-user))))
	    (setq passwd (or imap-password
			     (imap-read-passwd
			      (concat "IMAP password for " user "@" 
				      imap-server ": "))))
	    (when (and user passwd)
	      (if (funcall loginfunc user passwd)
		  (progn
		    (setq ret t
			  imap-username user)
		    (if (and (not imap-password)
			     (y-or-n-p "Store password for this session? "))
			(setq imap-password passwd)))
		(message "Login failed...")
		(setq passwd nil)
		(sit-for 1))))
	(quit (with-current-buffer buffer
		(setq user nil
		      passwd nil))))
      ret)))

(defun imap-kerberos4a-p (buffer)
  (imap-capability 'AUTH=KERBEROS_V4 buffer))

(defun imap-kerberos4-auth (buffer)
  (eq imap-stream 'kerberos4))

(defun imap-cram-md5-p (buffer)
  (imap-capability 'AUTH=CRAM-MD5 buffer))

(defun imap-cram-md5-auth (buffer)
  "Login to server using the AUTH CRAM-MD5 method."
  (imap-interactive-login
   buffer
   (lambda (user passwd)
     (imap-ok-p
      (imap-send-command-wait
       (list
	"AUTHENTICATE CRAM-MD5"
	(lambda (challenge)
	  (let* ((decoded (base64-decode challenge))
		 (hash (rfc2104-hash 'md5 64 16 passwd decoded))
		 (response (concat user " " hash))
		 (encoded (base64-encode response)))
	    encoded))))))))

(defun imap-login-p (buffer)
  (not (imap-capability 'X-LOGIN-CMD-DISABLED buffer)))

(defun imap-login-auth (buffer)
  "Login to server using the LOGIN command."
  (imap-interactive-login buffer 
			  (lambda (user passwd)
			    (imap-ok-p (imap-send-command-wait 
					(concat "LOGIN \"" user "\" \"" 
						passwd "\""))))))

(defun imap-anonymous-p (buffer)
  t)

(defun imap-anonymous-open (buffer)
  (with-current-buffer buffer
    (imap-ok-p (imap-send-command-wait
		(concat "LOGIN anonymous \"" (concat (user-login-name) "@" 
						     (system-name)) "\"")))))

;; Server functions:

(defun imap-open-1 (buffer)
  (with-current-buffer buffer
    (setq imap-current-mailbox nil
	  imap-current-message nil
	  imap-state 'initial
	  imap-process (funcall (nth 2 (assq imap-stream imap-stream-alist))
				"imap" buffer imap-server imap-port))
    (when imap-process
      (set-process-filter imap-process 'imap-arrival-filter)
      (set-process-sentinel imap-process 'imap-sentinel)
      (while (eq imap-state 'initial)
	(message "Waiting for server response...")
	(accept-process-output imap-process 1))
      (message "Waiting for server response...done")
      imap-process)))

(defun imap-open (server &optional port stream auth buffer)
  (setq buffer (or buffer (format " *imap* %s:%d" server (or port 0))))
  (unless (get-buffer buffer)
    (setq imap-buffer-list (cons buffer imap-buffer-list)))
  (with-current-buffer (get-buffer-create buffer)
    (let (stream-changed)
      (if (imap-opened buffer)
	  (imap-close buffer)
	(mapc 'make-variable-buffer-local imap-local-variables)
	(buffer-disable-undo))
      (if server (setq imap-server server))
      (if port (setq imap-port port))
      (if auth (setq imap-auth auth))
      (if stream (setq imap-stream stream))
      (if (and imap-stream imap-auth)
	  (imap-open-1 buffer)
	;; Determine network stream and/or auth mechanism to use
	(if (not (let ((imap-stream (or imap-stream imap-default-stream)))
		   (imap-open-1 buffer)))
	    ;; clean up and exit
	    (when (memq buffer imap-buffer-list)
	      (kill-buffer buffer)
	      (setq imap-buffer-list (delq buffer imap-buffer-list)))
	  ;; Choose stream.
	  (when (null imap-stream)
	    (let ((streams imap-streams))
	      (while (setq stream (pop streams))
		(if (funcall (nth 1 (assq stream imap-stream-alist)) buffer)
		    (setq stream-changed (not (eq (or imap-stream 
						      imap-default-stream)
						  stream))
			  imap-stream stream
			  streams nil)))
	      (unless imap-stream
		(error "Couldn't figure out a stream for server"))))
	  ;; Choose authenticator
	  (when (null imap-auth)
	    (let ((auths imap-authenticators))
	      (while (setq auth (pop auths))
		(if (funcall (nth 1 (assq auth imap-authenticator-alist)) 
			     buffer)
		    (setq imap-auth auth
			  auths nil)))
	      (unless imap-auth
		(error "Couldn't figure out authenticator for server"))))
	  (when stream-changed
	    (message "Reconnecting with %s..." imap-stream)
	    (imap-close buffer)
	    (imap-open-1 buffer))))))
  (if (imap-opened buffer)
      buffer))

(defun imap-opened (&optional buffer)
  (and (setq buffer (get-buffer (or buffer (current-buffer))))
       (buffer-live-p buffer)
       (with-current-buffer buffer
	 (and imap-process
	      (memq (process-status imap-process) '(open run))))))

(defun imap-authenticate (buffer &optional user passwd)
  (with-current-buffer buffer
    (when (eq imap-state 'nonauth)
      (make-variable-buffer-local 'imap-username)
      (make-variable-buffer-local 'imap-password)
      (if user (setq imap-username user))
      (if passwd (setq imap-password passwd))
      (if (funcall (nth 2 (assq imap-auth imap-authenticator-alist)) buffer)
	  (setq imap-state 'auth)))))

(defun imap-close (&optional buffer)
  "Close connection to server in BUFFER. If BUFFER is nil, the current
buffer is used."
  (with-current-buffer (or buffer (current-buffer))
    (and (imap-opened)
	 (not (imap-ok-p (imap-send-command-wait "LOGOUT")))
	 (error "Server %s didn't let me log out" imap-server))
    (when (and imap-process
	       (memq (process-status imap-process) '(open run)))
      (delete-process imap-process))
    (setq imap-current-mailbox nil
	  imap-current-message nil
	  imap-process nil)
    (erase-buffer)
    t))

(defun imap-capability (&optional identifier buffer)
  "Return a list of identifiers which the server support. If IDENTIFIER,
return non-nil if it's among the servers capabilities."
  (with-current-buffer (or buffer (current-buffer))
    (unless imap-capability
      (unless (imap-ok-p (imap-send-command-wait "CAPABILITY"))
	(setq imap-capability '(IMAP2))))
    (if identifier
	(memq identifier imap-capability)
      imap-capability)))

(defun imap-namespace (&optional buffer)
  "Return server's namespace."
  (with-current-buffer (or buffer (current-buffer))
    (unless imap-namespace
      (when (imap-capability 'NAMESPACE)
	(imap-send-command-wait 'NAMESPACE)))
    imap-namespace))


;; Mailbox functions:

(defun imap-mailbox-put (propname value &optional mailbox buffer)
  (with-current-buffer (or buffer (current-buffer))
    (put (intern (or mailbox imap-current-mailbox) imap-mailbox-data)
	 propname value)
    t))

(defun imap-mailbox-get (propname &optional mailbox buffer)
  (with-current-buffer (or buffer (current-buffer))
    (get (intern-soft (or mailbox imap-current-mailbox) imap-mailbox-data)
	 propname)))

(defun imap-mailbox-map (func &optional buffer)
  "Map a function across each mailbox in `imap-mailbox-data',
returning a list."
  (with-current-buffer (or buffer (current-buffer))
    (let (result)
      (mapatoms 
       (lambda (s)
	 (push (funcall func (symbol-name s)) result))
       imap-mailbox-data)
      result)))

(defun imap-mailbox-select (mailbox &optional buffer examine)
  (with-current-buffer (or buffer (current-buffer))
    (unless (and (string= mailbox imap-current-mailbox)
		 (or (and examine
			  (eq imap-state 'examine))
		     (and (not examine)
			  (eq imap-state 'selected))))
      (setq imap-current-mailbox mailbox)
      (if (imap-ok-p (imap-send-command-wait
		      (concat (if examine "EXAMINE" "SELECT") " " mailbox)))
	  (setq imap-message-data (make-vector imap-message-prime 0)
		imap-state (if examine 'examine 'selected))
	;; Failed SELECT unselects the current group
	(setq imap-current-mailbox nil
	      imap-message-data nil
	      imap-state 'auth)))
    imap-current-mailbox))

(defun imap-mailbox-unselect (&optional buffer)
  "Close current folder in BUFFER, without expunging articles."
  (with-current-buffer (or buffer (current-buffer))
    (when (or (eq imap-state 'auth)
	      (and (imap-capability 'UNSELECT)
		   (imap-ok-p (imap-send-command-wait "UNSELECT")))
	      (and (imap-ok-p 
		    (imap-send-command-wait (concat "EXAMINE "
						    imap-current-mailbox)))
		   (imap-ok-p (imap-send-command-wait "CLOSE"))))
      (setq imap-current-mailbox nil
	    imap-message-data nil
	    imap-state 'auth)
      t)))

(defun imap-mailbox-close (&optional buffer)
  "Expunge articles and close current folder in BUFFER."
  (with-current-buffer (or buffer (current-buffer))
    (when (imap-ok-p (imap-send-command-wait "CLOSE"))
      (setq imap-current-mailbox nil
	    imap-message-data nil
	    imap-state 'auth)
      t)))

(defun imap-mailbox-lsub (&optional buffer reference)
  "Clear the mailbox data and fill it with subscribed mailboxes on
server in BUFFER. REFERENCE is the implementation-specific string that
has to be passed to LSUB."
  (with-current-buffer (or buffer (current-buffer))
    (setq imap-mailbox-data (make-vector imap-mailbox-prime 0))
    (when (imap-ok-p (imap-send-command-wait 
		      (concat "LSUB \"" reference "\" \"*\"")))
      (imap-mailbox-map 'identity))))

(defun imap-mailbox-list (&optional buffer root have-delimiter reference)
  "List all mailboxes that starts with ROOT in BUFFER. If
HAVE-DELIMITER is non-nil, a hierarchy delimiter is not added to
ROOT. REFERENCE is the implementation-specific string that has to be
passed to LIST."
  (with-current-buffer (or buffer (current-buffer))
    (setq imap-mailbox-data (make-vector imap-mailbox-prime 0))
    ;; Find hierarchy separator if root turns out to be a mailbox
    ;; instead of just a prefix.
    (when root
      (unless have-delimiter
	(imap-send-command-wait (concat "LIST \"" reference "\" \"" root "\""))))
    (when (imap-ok-p 
	   (imap-send-command-wait 
	    (concat "LIST \"" reference "\" \"" root
		    (when (and root (not have-delimiter))
		      (imap-mailbox-get 'delimiter root))
		    "%\"")))
      (imap-mailbox-map 'identity))))

(defun imap-mailbox-subscribe (mailbox &optional buffer)
  "Send the SUBSCRIBE command on the mailbox to server in
BUFFER. Returns non-nil if successful."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p (imap-send-command-wait (concat "SUBSCRIBE " mailbox)))))

(defun imap-mailbox-unsubscribe (mailbox &optional buffer)
  "Send the SUBSCRIBE command on the mailbox to server in
BUFFER. Returns non-nil if successful."
  (with-current-buffer (or buffer (current-buffer))
    (imap-ok-p (imap-send-command-wait (concat "UNSUBSCRIBE " mailbox)))))


;; Message functions:

(defun imap-message-put (uid propname value &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (put (intern (number-to-string uid) imap-message-data)
	 propname value)
    t))

(defun imap-message-get (uid propname &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (get (intern-soft (number-to-string uid) imap-message-data)
	 propname)))

(defun imap-message-map (func propname &optional buffer)
  "Map a function across each mailbox in `imap-message-data',
returning a list."
  (with-current-buffer (or buffer (current-buffer))
    (let (result)
      (mapatoms
       (lambda (s)
	 (push (funcall func (get s 'UID) (get s propname)) result))
       imap-message-data)
      result)))

(defun imap-search (predicate &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (imap-mailbox-put 'search nil)
    (when (imap-ok-p (imap-send-command-wait (concat "UID SEARCH " predicate)))
      (imap-mailbox-get 'search))))

(defun imap-message-flags-set (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " FLAGS" (if silent ".SILENT") " (" flags ")"))))))

(defun imap-message-flags-del (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " -FLAGS" (if silent ".SILENT") " (" flags ")"))))))

(defun imap-message-flags-add (articles flags &optional buffer silent)
  (when (and articles flags)
    (with-current-buffer (or buffer (current-buffer))
      (imap-ok-p (imap-send-command-wait
		  (concat "UID STORE " articles
			  " +FLAGS" (if silent ".SILENT") " (" flags ")"))))))


;; Internal functions.

(defun imap-send-command (command &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (if (not (listp command)) (setq command (list command)))
    (let ((tag (setq imap-tag (1+ imap-tag)))
	  cmdstr cmd)
      (setq cmdstr (concat (number-to-string imap-tag) " "))
      (while (setq cmd (pop command))
	(cond ((stringp cmd)
	       (setq cmdstr (concat cmdstr cmd)))
	      (t
	       (error "Unknown command type"))))
      (setq cmdstr (concat cmdstr imap-client-eol))
      (and imap-log
	   (with-current-buffer (get-buffer-create imap-log)
	     (goto-char (point-max))
	     (insert cmdstr)))
      (process-send-string imap-process cmdstr)
      tag)))

(defun imap-wait-for-tag (tag &optional buffer)
  (with-current-buffer (or buffer (current-buffer))
    (while (< imap-reached-tag tag)
      (accept-process-output imap-process))))

(defun imap-ok-p (&rest foo)
  t)

(defun imap-send-command-wait (command &optional buffer)
  (imap-wait-for-tag (imap-send-command command buffer) buffer))

(defun imap-sentinel (process string)
  (delete-process process))

(defun imap-find-next-line ()
  "Return point at end of current line, taking into account
literals. Return nil if no complete line has arrived."
  (when (re-search-forward (concat imap-server-eol "\\|{\\([0-9]+\\)}" imap-server-eol)
			   nil t)
    (if (match-string 1)
	(if (< (point-max) (+ (point) (string-to-number (match-string 1))))
	    nil
	  (goto-char (+ (point) (string-to-number (match-string 1))))
	  (imap-find-next-line))
      (point))))

(defun imap-arrival-filter (proc string)
  "IMAP process filter."
  (with-current-buffer (process-buffer proc)
    (goto-char (point-max))
    (insert string)
    (and imap-log
	 (with-current-buffer (get-buffer-create imap-log)
	   (goto-char (point-max))
	   (insert string)))
    (goto-char (point-min))
    (let (end)
      (while (setq end (imap-find-next-line))
	(save-restriction
	  (narrow-to-region (point-min) end)
	  (goto-char (point-min))
	  ;; unwind-protect when parser is debugged
	  (cond ((eq imap-state 'initial)
		 (imap-parse-greeting))
		((or (eq imap-state 'auth)
		     (eq imap-state 'nonauth)
		     (eq imap-state 'selected))
		 (imap-parse-response))
		(t
		 (error "Unknown state %s in arrival filter" imap-state)))
	  (delete-region (point-min) (point-max)))))))


;; Imap parser.

;;   greeting        = "*" SP (resp-cond-auth / resp-cond-bye) CRLF
;;
;;   resp-cond-auth  = ("OK" / "PREAUTH") SP resp-text
;;                       ; Authentication condition
;;
;;   resp-cond-bye   = "BYE" SP resp-text

(defun imap-parse-greeting ()
  "Parse a IMAP greeting."
  (cond ((looking-at "\\* OK ")
	 (setq imap-state 'nonauth))
	((looking-at "\\* PREAUTH ")
	 (setq imap-state 'auth))
	((looking-at "\\* BYE ")
	 (setq imap-state 'closed))))

;;   response        = *(continue-req / response-data) response-done
;;
;;   continue-req    = "+" SP (resp-text / base64) CRLF
;;
;;   response-data   = "*" SP (resp-cond-state / resp-cond-bye /
;;                     mailbox-data / message-data / capability-data) CRLF
;;
;;   response-done   = response-tagged / response-fatal
;;
;;   response-fatal  = "*" SP resp-cond-bye CRLF
;;                       ; Server closes connection immediately
;;
;;   response-tagged = tag SP resp-cond-state CRLF

(defun imap-parse-response ()
  "Parse a IMAP command response."
  (let ((token (read (current-buffer))))
    (cond ((eq token '*)
	   (let* ((response (read (current-buffer)))
		  (func (cdr (assq response imap-parse-response-data-cb))))
	     (forward-char)
	     (when (integerp response)
	       (setq func (cdr (assq (read (current-buffer))
				     imap-parse-response-data-cb)))
	       (forward-char))
	     (if func
		 (funcall func response)
	       (message "Unknown untagged response: %s" response))))
	  ((integerp token)
	   (let ((status (read (current-buffer))))
	     (cond ((eq status 'OK)
		    (setq imap-reached-tag (max imap-reached-tag token)))
		   ((eq status 'NO)
		    (error "Imap server said no: %s" 
			   (buffer-substring (point) (point-max))))
		   ((eq status 'BAD)
		    (error "Internal protocol error: %s"
			   (buffer-substring (point) (point-max))))
		   (t
		    (error "Unknown tagged status: %s" status)))))
	  ((eq token '+)
	   (imap-parse-continue-req))
	  (t
	   (error "Unknown symbol error")))))

;;   resp-text-code  = "ALERT" /
;;                     "BADCHARSET [SP "(" astring *(SP astring) ")" ] /
;;                     "NEWNAME" SP string SP string / 
;;		       "PARSE" /
;;                     "PERMANENTFLAGS" SP "(" [flag-perm *(SP flag-perm)] ")" /
;;                     "READ-ONLY" / 
;;		       "READ-WRITE" / 
;;	 	       "TRYCREATE" /
;;                     "UIDNEXT" SP nz-number / 
;;		       "UIDVALIDITY" SP nz-number /
;;                     "UNSEEN" SP nz-number /
;;                     resp-text-atom [SP 1*<any TEXT-CHAR except "]">]
;;
;;   flag-perm       = flag / "\*"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   flag-keyword    = atom
;;
;;   resp-text-atom  = 1*<any ATOM-CHAR except "]">

(defun imap-response-data-text-code (response)
  (assert (eq (char-after) ?\[))
  (forward-char)
  (cond ((search-forward "PERMANENTFLAGS " nil t)
	 (imap-mailbox-put 'permanentflags (imap-parse-flag-list)))
	((search-forward "READ-ONLY" nil t)
	 (imap-mailbox-put 'read-only t))
	((search-forward "UIDNEXT " nil t)
	 (imap-mailbox-put 'uidnext (read (current-buffer))))
	((looking-at "UIDVALIDITY \\([0-9]+\\)")
	 (imap-mailbox-put 'uidvalidity (match-string 1)))
	((search-forward "UNSEEN " nil t)
	 (imap-mailbox-put 'unseen (read (current-buffer))))
	((search-forward "NEWNAME " nil t)
	 (let (oldname newname)
	   (setq oldname (imap-parse-string))
	   (forward-char)
	   (setq newname (imap-parse-string))
	   (imap-mailbox-put 'newname newname oldname)))
	((search-forward "TRYCREATE" nil t)
	 (imap-mailbox-put 'trycreate t))
	((search-forward "ALERT] " nil t)
	 (message "Imap server %s information: %s"
		  imap-server
		  (buffer-substring (point)
				    (- (point-max) (length imap-server-eol)))))
	(t
	 (error "Unknown response code: %s" (read (current-buffer))))))

;;   response-data   = "*" SP (resp-cond-state / resp-cond-bye /
;;                     mailbox-data / message-data / capability-data) CRLF
;;
;;   resp-cond-bye   = "BYE" SP resp-text

(defun imap-response-data-bye (response)
  t)

;;   response-data   = "*" SP (resp-cond-state / resp-cond-bye /
;;                     mailbox-data / message-data / capability-data) CRLF
;;
;;   resp-cond-state = ("OK" / "NO" / "BAD") SP resp-text
;;                       ; Status condition
;;
;;   resp-cond-bye   = "BYE" SP resp-text
;;
;;   mailbox-data    =  "FLAGS" SP flag-list /
;;		        "LIST" SP mailbox-list /
;;                      "LSUB" SP mailbox-list /
;;		        "SEARCH" *(SP nz-number) /
;;                      "STATUS" SP mailbox SP "("
;;	                      [status-att SP number *(SP status-att SP number)] ")" /
;;                      number SP "EXISTS" /
;;		        number SP "RECENT"
;;
;;   message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
;;
;;   capability      = "AUTH=" auth-type / atom
;;                       ; New capabilities MUST begin with "X" or be
;;                       ; registered with IANA as standard or
;;                       ; standards-track
;;
;;   capability-data = "CAPABILITY" *(SP capability) SP "IMAP4rev1"
;;                     *(SP capability)
;;                       ; IMAP4rev1 servers which offer RFC 1730
;;                       ; compatibility MUST list "IMAP4" as the first
;;                       ; capability.

(defun imap-response-data-exists (response)
  (imap-mailbox-put 'exists response))

(defun imap-response-data-expunge (response)
  (imap-mailbox-put 'exists response))

(defun imap-response-data-recent (response)
  (imap-mailbox-put 'recent response))

(defun imap-response-data-capability (response)
  (and (looking-at (concat "[^" imap-server-eol "]+" imap-server-eol))
       (setq imap-capability (read (concat "(" (match-string 0) ")")))))

;;   mailbox         = "INBOX" / astring
;;                       ; INBOX is case-insensitive.  All case variants of
;;                       ; INBOX (e.g. "iNbOx") MUST be interpreted as INBOX
;;                       ; not as an astring.  An astring which consists of
;;                       ; the case-insensitive sequence "I" "N" "B" "O" "X"
;;                       ; is considered to be INBOX and not an astring.
;;                       ;  Refer to section 5.1 for further
;;                       ; semantic details of mailbox names.
;;
;;   mailbox-list    = "(" [mbx-list-flags] ")" SP
;;                      (DQUOTE QUOTED-CHAR DQUOTE / nil) SP mailbox
;;
;;   mbx-list-flags  = *(mbx-list-oflag SP) mbx-list-sflag
;;                     *(SP mbx-list-oflag) /
;;                     mbx-list-oflag *(SP mbx-list-oflag)
;;
;;   mbx-list-oflag  = "\Noinferiors" / flag-extension
;;                       ; Other flags; multiple possible per LIST response
;;
;;   mbx-list-sflag  = "\Noselect" / "\Marked" / "\Unmarked"
;;                       ; Selectability flags; only one per LIST response

(defun imap-response-data-list (type)
  (let (flags delimiter mailbox)
    (if (looking-at "\\(([^)]*)\\) \\(NIL\\|\"\\(.\\)\\\"\\) ")
	(setq flags (match-string 1)
	      delimiter (match-string 3))
      (error "Parse error"))
    (goto-char (match-end 0))
    (setq flags (read (replace-in-string flags "\\\\" "")))
    (if (looking-at "\"?\\([^\r\n\"]+\\)\"?")
	(setq mailbox (match-string 1))
      (error "Parse error"))
    (goto-char (match-end 0))
    (imap-mailbox-put type t mailbox)
    (imap-mailbox-put 'flags flags mailbox)
    (imap-mailbox-put 'delimiter delimiter mailbox)))

(defun imap-response-data-flags (response)
  (imap-mailbox-put 'flags (imap-parse-flag-list)))

;;   msg-att         = "(" (msg-att-dynamic / msg-att-static)
;;                      *(SP (msg-att-dynamic / msg-att-static)) ")"
;;
;;   msg-att-dynamic = "FLAGS" SP "(" [flag-fetch *(SP flag-fetch)] ")"
;;                       ; MAY change for a message
;;
;;   flag-fetch      = flag / "\Recent"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-keyword    = atom
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   msg-att-static  = "ENVELOPE" SP envelope / 
;;                     "INTERNALDATE" SP date-time /
;;                     "RFC822" [".HEADER" / ".TEXT"] SP nstring /
;;                     "RFC822.SIZE" SP number /
;;                     "BODY" ["STRUCTURE"] SP body /
;;                     "BODY" section ["<" number ">"] SP nstring /
;;                     "UID" SP uniqueid
;;                       ; MUST NOT change for a message
;;
;;   section         = "[" [section-spec] "]"
;;
;;   section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
;;                     "TEXT"
;;                       ; top-level or MESSAGE/RFC822 part
;;
;;   section-part    = nz-number *("." nz-number)
;;                       ; body part nesting
;;
;;   section-spec    = section-msgtext / (section-part ["." section-text])
;;
;;   section-text    = section-msgtext / "MIME"
;;                       ; text other than actual body part (headers, etc.)
;;
;;   date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year
;;                     SP time SP zone DQUOTE
;;
;;   uniqueid        = nz-number
;;                       ; Strictly ascending

(defun imap-response-data-fetch (response)
  ;; we don't know where to store things until we know the UID.
  ;; close your eyes now...
  (save-excursion
    (unless (re-search-forward "UID \\([0-9]+\\)" nil t)
      (error "Can't find UID"))
    (setq imap-current-message (string-to-number (match-string 1))))
  ;; ...you may open them again.
  (imap-message-put imap-current-message 'UID imap-current-message)
  (assert (eq (char-after) ?\())
  (while (not (eq (char-after) ?\)))
    (forward-char)
    (let ((token (read (current-buffer))))
      (forward-char)
      (cond ((eq token 'UID)
	     (forward-sexp))
	    ((eq token 'FLAGS)
	     (imap-message-put imap-current-message 'FLAGS
			       (imap-parse-flag-list)))
	    ((eq token 'ENVELOPE)
	     (imap-message-put imap-current-message 'ENVELOPE
			       (imap-parse-envelope)))
	    ((eq token 'INTERNALDATE)
	     (imap-message-put imap-current-message 'INTERNALDATE
			       (read (current-buffer))))
	    ((eq token 'RFC822)
	     (imap-message-put imap-current-message 'RFC822
			       (imap-parse-nstring)))
	    ((eq token 'RFC822.HEADER)
	     (imap-message-put imap-current-message 'RFC822.HEADER
			       (imap-parse-nstring)))
	    ((eq token 'RFC822.TEXT)
	     (imap-message-put imap-current-message 'RFC822.TEXT
			       (imap-parse-nstring)))
	    ((eq token 'RFC822.SIZE)
	     (imap-message-put imap-current-message 'RFC822.SIZE
			       (read (current-buffer))))
	    ((eq token 'BODY)
	     (imap-message-put imap-current-message 'BODY
			       (imap-parse-body)))
	    ((eq token 'BODYSTRUCTURE)
	     (imap-message-put imap-current-message 'BODYSTRUCTURE
			       (imap-parse-body)))
	    (t
	     (error "Unknown message data: %s" token))))))

(defun imap-response-data-search (response)
  (and (looking-at (concat "[^" imap-server-eol "]+" imap-server-eol))
       (imap-mailbox-put 'search (read (concat "(" (match-string 0) ")")))))

(defun imap-response-data-status (response)
  (assert (eq (char-after) ?\())
  (while (not (eq (char-after) ?\)))
    (forward-char)
    (let ((token (read (current-buffer))))
      (forward-char)
      (cond ((eq token 'MESSAGES)
	     (imap-mailbox-put 'MESSAGES (read (current-buffer))))
	    ((eq token 'RECENT)
	     (imap-mailbox-put 'RECENT (read (current-buffer))))
	    ((eq token 'UIDNEXT)
	     (imap-mailbox-put 'UIDNEXT (read (current-buffer))))
	    ((eq token 'UIDVALIDITY)
	     (and (looking-at "[0-9]+")
		  (imap-mailbox-put 'UIDVALIDITY (match-string 0))
		  (goto-char (match-end 0))))
	    ((eq token 'UNSEEN)
	     (imap-mailbox-put 'UNSEEN (read (current-buffer))))
	     (t
	      (error "Unknown status data: %s" token))))))

;;   flag-list       = "(" [flag *(SP flag)] ")"
;;
;;   flag            = "\Answered" / "\Flagged" / "\Deleted" /
;;                     "\Seen" / "\Draft" / flag-keyword / flag-extension
;;                       ; Does not include "\Recent"
;;
;;   flag-keyword    = atom
;;
;;   flag-extension  = "\" atom
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept flag-extension flags.  Server
;;                       ; implementations MUST NOT generate
;;                       ; flag-extension flags except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.

(defun imap-parse-flag-list ()
  (when (looking-at "([^)]*)")
    (goto-char (match-end 0))
    (read (replace-in-string (match-string 0) "\\\\" "\\\\\\\\"))))

;;   envelope        = "(" env-date SP env-subject SP env-from SP env-sender SP
;;                     env-reply-to SP env-to SP env-cc SP env-bcc SP
;;                     env-in-reply-to SP env-message-id ")"
;;
;;   env-bcc         = "(" 1*address ")" / nil
;;
;;   env-cc          = "(" 1*address ")" / nil
;;
;;   env-date        = nstring
;;
;;   env-from        = "(" 1*address ")" / nil
;;
;;   env-in-reply-to = nstring
;;
;;   env-message-id  = nstring
;;
;;   env-reply-to    = "(" 1*address ")" / nil
;;
;;   env-sender      = "(" 1*address ")" / nil
;;
;;   env-subject     = nstring
;;
;;   env-to          = "(" 1*address ")" / nil

(defun imap-parse-envelope ()
  ;; xxx: does not handle literals
  (read (current-buffer)))

;;   body            = "(" body-type-1part / body-type-mpart ")"
;;
;;   body-extension  = nstring / number /
;;                      "(" body-extension *(SP body-extension) ")"
;;                       ; Future expansion.  Client implementations
;;                       ; MUST accept body-extension fields.  Server
;;                       ; implementations MUST NOT generate
;;                       ; body-extension fields except as defined by
;;                       ; future standard or standards-track
;;                       ; revisions of this specification.
;;
;;   body-ext-1part  = body-fld-md5 [SP body-fld-dsp [SP body-fld-lang
;;                     *(SP body-extension)]]
;;                       ; MUST NOT be returned on non-extensible
;;                       ; "BODY" fetch
;;
;;   body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
;;                     *(SP body-extension)]]
;;                       ; MUST NOT be returned on non-extensible
;;                       ; "BODY" fetch
;;
;;   body-fields     = body-fld-param SP body-fld-id SP body-fld-desc SP
;;                     body-fld-enc SP body-fld-octets
;;
;;   body-fld-desc   = nstring
;;
;;   body-fld-dsp    = "(" string SP body-fld-param ")" / nil
;;
;;   body-fld-enc    = (DQUOTE ("7BIT" / "8BIT" / "BINARY" / "BASE64"/
;;                     "QUOTED-PRINTABLE") DQUOTE) / string
;;
;;   body-fld-id     = nstring
;;
;;   body-fld-lang   = nstring / "(" string *(SP string) ")"
;;
;;   body-fld-lines  = number
;;
;;   body-fld-md5    = nstring
;;
;;   body-fld-octets = number
;;
;;   body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
;;
;;   body-type-1part = (body-type-basic / body-type-msg / body-type-text)
;;                     [SP body-ext-1part]
;;
;;   body-type-basic = media-basic SP body-fields
;;                       ; MESSAGE subtype MUST NOT be "RFC822"
;;
;;   body-type-mpart = 1*body SP media-subtype
;;                     [SP body-ext-mpart]
;;
;;   body-type-msg   = media-message SP body-fields SP envelope
;;                     SP body SP body-fld-lines
;;
;;   body-type-text  = media-text SP body-fields SP body-fld-lines
;;
;;   media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" / "MESSAGE" /
;;                     "VIDEO") DQUOTE) / string) SP media-subtype
;;                       ; Defined in [MIME-IMT]

;; ("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 7415 213)
;; (("TEXT" "PLAIN" ("CHARSET" "US-ASCII") NIL NIL "7BIT" 9 2)(("TEXT" "PLAIN" ("CHARSET" "US-ASCII") NIL NIL "7BIT" 8 2)("IMAGE" "GIF" NIL NIL NIL "BASE64" 2802) "ALTERNATIVE")("TEXT" "PLAIN" NIL NIL NIL "7BIT" 7 1) "MIXED")
;; (("IMAGE" "JPEG" NIL NIL NIL "BASE64" 1222)("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 2 1)(("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 21 2)("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 20 2)(("IMAGE" "JPEG" NIL NIL NIL "BASE64" 1222) "MIXED") "ALTERNATIVE")("TEXT" "PLAIN" ("CHARSET" "us-ascii") NIL NIL "7BIT" 111 5) "MIXED")

(defun imap-parse-body ()
  ;; xxx: does not handle literals
  (read (current-buffer)))

;;   atom            = 1*ATOM-CHAR
;;
;;   ATOM-CHAR       = <any CHAR except atom-specials>
;;
;;   atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
;;                     quoted-specials
;;
;;   list-wildcards  = "%" / "*"
;;
;;   quoted-specials = DQUOTE / "\"

(defun imap-parse-atom ()
  (and (looking-at "[^(){ %*\"\\\r\n]+") ;; xxx: CTL
       (intern (match-string 0))))

;;   astring         = atom / string

(defun imap-parse-astring ()
  (or (imap-parse-atom)
      (imap-parse-string)))

;;   nil             = "NIL"
;;
;;   nstring         = string / nil

(defun imap-parse-nstring ()
  (let ((str (imap-parse-string)))
    (if (string= "NIL" str)
	nil
      str)))

;;   string          = quoted / literal

(defun imap-parse-string ()
  (or (imap-parse-quoted)
      (imap-parse-literal)))
  
;;   quoted          = DQUOTE *QUOTED-CHAR DQUOTE
;;
;;   QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> /
;;                     "\" quoted-specials
;;
;;   quoted-specials = DQUOTE / "\"
;;
;;   TEXT-CHAR       = <any CHAR except CR and LF>

(defun imap-parse-quoted ()
  (and (looking-at "\"\\([^\r\n]*\\)\"")
       (match-string 1)))

;;   literal         = "{" number "}" CRLF *CHAR8
;;                       ; Number represents the number of CHAR8s

(defun imap-parse-literal ()
  (when (looking-at "{\\([0-9]+\\)}\r\n")
    (let ((pos (match-end 0))
	  (len (string-to-number (match-string 1))))
      (if (< (point-max) (+ pos len))
	  nil
	(goto-char (+ pos len))
	(buffer-substring pos (+ pos len))))))


;; Utility functions.

(defun imap-read-passwd (prompt &rest args)
  "Read a password using PROMPT. If ARGS, PROMPT is used as an
argument to `format'."
  (let ((prompt (if args
		    (apply 'format prompt args)
		  prompt)))
    (funcall (if (load "passwd" t)
		 'read-passwd
	       (unless (fboundp 'ange-ftp-read-passwd)
		 (autoload 'ange-ftp-read-passwd "ange-ftp"))
	       'ange-ftp-read-passwd) prompt)))

(when imap-debug ; (untrace-all)
  (require 'trace)
  (buffer-disable-undo (get-buffer-create imap-debug))
  (mapc (lambda (f) (trace-function-background f imap-debug)) 
        '(
imap-kerberos4s-p
imap-kerberos4-open
imap-ssl-p
imap-ssl-open-1
imap-ssl-open
imap-network-p
imap-network-open
imap-interactive-login
imap-kerberos4a-p
imap-kerberos4-auth
imap-cram-md5-p
imap-cram-md5-auth
imap-login-p
imap-login-auth
imap-anonymous-p
imap-anonymous-open
imap-open-1
imap-open
imap-opened
imap-authenticate
imap-close
imap-capability
imap-namespace
imap-mailbox-put
imap-mailbox-get
imap-mailbox-map
imap-mailbox-select
imap-mailbox-unselect
imap-mailbox-close
imap-mailbox-lsub
imap-mailbox-list
imap-mailbox-subscribe
imap-mailbox-unsubscribe
imap-message-put
imap-message-get
imap-message-map
imap-search
imap-message-flags-set
imap-message-flags-del
imap-message-flags-add
imap-send-command
imap-wait-for-tag
imap-ok-p
imap-send-command-wait
imap-sentinel
imap-find-next-line
imap-arrival-filter
imap-parse-greeting
imap-parse-response
imap-response-data-text-code
imap-response-data-bye
imap-response-data-exists
imap-response-data-expunge
imap-response-data-recent
imap-response-data-capability
imap-response-data-list
imap-response-data-flags
imap-response-data-fetch
imap-response-data-search
imap-response-data-status
imap-parse-base64
imap-parse-continue-req
imap-parse-flag-list
imap-parse-envelope
imap-parse-body
imap-parse-atom
imap-parse-astring
imap-parse-nstring
imap-parse-string
imap-parse-quoted
imap-parse-literal
imap-parse-text
imap-read-passwd
	  )))
	
(provide 'imap)
